

set -ex



pip check
export OPENBLAS_NUM_THREADS=1
export OMP_NUM_THREADS=1
pytest --timeout 300 -n 1 --verbose --pyargs sklearn -k "not (_not_a_real_test or test_loadings_converges or test_ridge_regression_check_arguments_validity)"
exit 0
